#include <GLUT/glut.h>									//GLUT library
#include "UnstructuredGrid3D.h"							//the 3D unstructured mesh class
#include "MeshRenderer.h"								//a simple renderer for 3D unstructured grids, demonstrating flat/smooth shading
#include "PlyReader.h"
#include "zpr.h"
#include <iostream>
#include <vector>
#include <stdlib.h>

using namespace std;



MeshRenderer		renderer;							//Renderer to draw the meshes
UnstructuredGrid3D*	mesh			=  0;				//Original mesh
UnstructuredGrid3D*	smooth_grid		=  0;				//Smoothed and/or noised mesh
vector<Point3d>		smoothed_points;					//Help array for the smoothing

int					drawing_style   =  MeshRenderer::DRAW_C1_CELLS;
														//Draw mesh as shaded using vertex normals
float				delta			=  0.4;				//Smoothing factor for shrinking 
														//(see G. Taubin, "Curve and surface smoothing without shrinkage", Proc. ICCV'95)



static void mouseclick(int button, int state, int x, int y)
{
	int keys = glutGetModifiers();						//The mouse events are interpreted as follows:	
	if (keys & GLUT_ACTIVE_CTRL)						// 
	button = GLUT_MIDDLE_BUTTON;						// -left button + move:                         rotate viewpoint
	if (keys & GLUT_ACTIVE_SHIFT)						// -middle button (or left button+Control key): zoom in / out
	button = GLUT_RIGHT_BUTTON;							// -right buttom (or left button+Shift key):    translate viewpoint

	zprMouse(button,state,x,y);							//Use the ZPR library to manipulate the viewpoint. The library sets the modelview  
														//OpenGL matrix based on the mouse events, thereby changing the current viewpoint.
}



static void mousemotion(int x, int y)
{
	zprMotion(x,y);
	glutPostRedisplay();
}


static void viewing(int width, int height)					//GLUT window-reshape callback
{
	glutReshapeWindow(width,height);  
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(60,float(width)/height,-100,100);
	glViewport(0,0,width,height);
}


static void addNoise(UnstructuredGrid3D* mesh,float amplitude)
{															//Add noise to the mesh vertices in the direction of the vertex normals.
															//The noise ranges in [0,amplitude]
	VectorAttributes&	normals = mesh->getPointNormals();
	
	for(int i=0;i<mesh->numPoints();++i)
	{
		Point3d point;
		mesh->getPoint(i,point.data);
		
		Point3d n = normals.getC0Vector(i);					//Get vertex normal															
		float dist = amplitude*((float(rand())/RAND_MAX-0.5)/2);
															//Compute amount of normal shift for this vertex
		point += n*dist;									//Shift the vertex along the normal	
		mesh->setPoint(i,point.data);						//Copy shifted vertex back into the mesh
	}	
		
    mesh->computeFaceNormals();								//Recompute normals since we changed vertices		
    mesh->computeVertexNormals();
}


static void smooth(UnstructuredGrid3D* mesh,float delta)	//Compute one iteration of Laplacian smoothing of 'mesh'.	
{															//'delta' ([0,1]) gives the amount of smoothing
	for(int i=0;i<mesh->numPoints();++i)
	{
		mesh->getPoint(i,smoothed_points[i].data);

		const UnstructuredGrid3D::CellsOfPoint& c = mesh->cellsOfPoint(i);

		Point3d average;	
		int     N=0;
		for(UnstructuredGrid3D::CellsOfPoint::const_iterator it=c.begin();it!=c.end();++it)
		{
		   int cj = *it;
		   int cv[3];
		   mesh->getCell(cj,cv); 
		   for(int j=0;j<3;++j)		
		   {
		       if (cv[j]==i) continue;
				
			   Point3d v;
			   mesh->getPoint(cv[j],v.data);
			   average += v;
			   ++N;
		   }	
		}   

		if (N>0)
		{	
			average /= N;
			average -= smoothed_points[i];
			average *= delta;
			smoothed_points[i] += average;	
		}		
	}
	
	for(int i=0;i<mesh->numPoints();++i)
		mesh->setPoint(i,smoothed_points[i].data);
}


static void smoothMesh(UnstructuredGrid3D* mesh, float delta, float mu)
{															//Taubin smoothing:
	if (delta) smooth(mesh,delta);							//1. Apply a shrinking iteration
 	if (mu) smooth(mesh,mu);								//2. Apply an inflation iteration (to compensate for volume loss)
    mesh->computeFaceNormals();								//3. Recompute normals since we changed the vertices	
    mesh->computeVertexNormals();
}


static void keyboard(unsigned char ch,int,int)				//GLUT keyboard callback
{
	switch(ch)
	{
	case ' ':												//Do one smoothing iteration
		{
		  smoothMesh(smooth_grid,delta,-delta+0.1);
		  break;
		}  
	case '+':												//Increase the amount of smoothing
	case '=':
	    {
		  delta += 0.1; if (delta>1) delta=1;
		  cout<<"Smoothing factor: "<<delta<<endl;
		  break;
		}  
	case '-':												//Decrease the amount of smoothing
	case '_':
	    {
		  delta -= 0.1; if (delta<0.1) delta=0.1;
		  cout<<"Smoothing factor: "<<delta<<endl;
		  break;
		}  
	case 'r':												//Reset the mesh to the original one
	case 'R':
		{
		  *smooth_grid = *mesh;
		  break;
		}				
	case 'n':												//Add a small amount of normal noise.
	case 'N':												//This allows testing next the effectiveness of smoothing
		{
		  addNoise(smooth_grid,0.01);
		  break;
		} 	
	case 'm':												//Toggle through the various drawing styles of the mesh renderer
		{
		  drawing_style = (++drawing_style) % 4;		
		  renderer.setDrawingStyle((MeshRenderer::DRAW_STYLE)drawing_style);
		  break;
		}
	}	
	
	glutPostRedisplay();	
}		


static void display()										//GLUT display callback
{
   renderer.draw(*smooth_grid);
}



int main(int argc,char* argv[])								//Main program
{
	cout<<"2. Laplacian mesh smoothing"<<endl;
	cout<<"In the graphics window, press"<<endl;
	cout<<"   space:  Do one smoothing iteration"<<endl;
	cout<<"   n:      Add a small amount of noise to the mesh"<<endl;
	cout<<"   +,-:    Increase/decrease the smoothing factor"<<endl;
	cout<<"   m:      Toggle through the various shading modes for the mesh"<<endl;   
	
	
	const char* filename = (argc>1)? argv[1]:"DATA/bunny.ply";
	
	glutInit(&argc, argv);									//1. Initialize the GLUT toolkit
	glutInitDisplayMode(GLUT_RGB|GLUT_DOUBLE|GLUT_DEPTH);	//2. Ask GLUT to create next windows with a RGB framebuffer and a Z-buffer too
	glutInitWindowSize(500,500);							//3. Tell GLUT how large are the windows we want to create next
	glutCreateWindow("2. Laplacian mesh smoothing");		


	PlyReader rdr;											//6.  Read a 3D mesh stored in a file in the PLY format
    mesh = rdr.read(filename);
    mesh->normalize();										//7.  Normalize the mesh in the [-1,1] cube. This makes setting the OpenGL projection easier.
    mesh->computeFaceNormals();								//8.  Compute face and vertex normals for the mesh. This allows us to shade the mesh next.
    mesh->computeVertexNormals();
	mesh->computePointTriangles();

	srand(time(0));


	smooth_grid = new UnstructuredGrid3D(*mesh);

	smoothed_points.resize(mesh->numPoints());

	zprInit(0,0,0);


	renderer.setDrawingStyle((MeshRenderer::DRAW_STYLE)drawing_style);
	   
	glutKeyboardFunc(keyboard);
	glutMouseFunc(mouseclick);
	glutMotionFunc(mousemotion);
	glutReshapeFunc(viewing);
	glutDisplayFunc(display);
	
	glutMainLoop();

	return 0;
}




